{****************************************************************************

                     Copyright (c) 1993,96 by Florian Klaempfl

 ****************************************************************************}

{ History:
      30th september 1996:
         + english comments (FPK)
         + _2pchar renamed to pstring2pchar (FPK)
         + _2pstring renamed to pchar2pstring (FPK)
      15th october 1996:
         + tstringcontainer is compilable
         + full compilable
}
unit cobjects;

  interface

    uses
       strings;

    type
       pstring = ^string;

       { some help data types }
       pstringitem = ^tstringitem;

       tstringitem = record
          data : pstring;
          next : pstringitem;
       end;

       plinkedlist_item = ^tlinkedlist_item;

       tlinkedlist_item = object
          next,last : plinkedlist_item;

          { does nothing }
          destructor done;virtual;
       end;

       plinkedlist = ^tlinkedlist;

       { this implements a double linked list }
       tlinkedlist = object
          first,last : plinkedlist_item;
          constructor init;
          destructor done;

          { disposes the items of the list }
          procedure clear;

          { concats a new item at the end }
          procedure concat(p : plinkedlist_item);

          { inserts a new item at the begin }
          procedure insert(p : plinkedlist_item);

          { inserts another list at the begin and make this list empty }
          procedure insertlist(p : plinkedlist);

          { concats another list at the end and make this list empty }
          procedure concatlist(p : plinkedlist);
       end;

       { string container }
       pstringcontainer = ^tstringcontainer;

       tstringcontainer = object
          root,last : pstringitem;

          { if this is set to true, doubles are allowed }
          { true is default                             }
          doubles : boolean;
          constructor init;
          destructor done;

          { inserts a string }
          procedure insert(const s : string);

          { gets a string }
          function get : string;

          { deletes all strings }
          procedure clear;
       end;

       pbufferedfile = ^tbufferedfile;

       { this is implemented to allow buffered binary I/O }
       { remark: input is currently not implemented       }
       tbufferedfile = object
           f : file;
           buf : pchar;
           bufsize : longint;
           bufpos : longint;

           { 0 output, 1 input }
           iomode : byte;

           { inits a buffer with the size bufsize which is assigned to }
           { the file  filename                                        }
           constructor init(const filename : string;_bufsize : longint);

           { close the file, if needed and releases the memory }
           destructor done;

           { opens the file for input, other accesses are rejected }
           procedure reset;

           { opens the file for output, other accesses are rejected }
           procedure rewrite;

           { reads or writes the buffer from or to disk }
           procedure flush;

           { writes a string to the file }
           { the string is written without a length byte }
           procedure write_string(const s : string);

           { writes a zero terminated string }
           procedure write_pchar(p : pchar);
       end;

    { releases the string p and assignes nil to p }
    { if p=nil then freemem isn't called          }
    procedure stringdispose(var p : pstring);

    { allocates mem for a copy of s, copies s to this mem and returns }
    { a pointer to this mem                                           }
    function stringdup(const s : string) : pstring;

    { allocates memory for s and copies s as zero terminated string
      to that mem and returns a pointer to that mem }
    function strpnew(const s : string) : pchar;

    { makes a char lowercase, with spanish, french and german char set }
    function lowercase(c : char) : char;

    { makes zero terminated string to a pascal string }
    { the data in p is modified and p is returned     }
    function pchar2pstring(p : pchar) : pstring;

    { ambivalent to pchar2pstring }
    function pstring2pchar(p : pstring) : pchar;

  implementation

    function pchar2pstring(p : pchar) : pstring;

      var
         w : word;
         i : longint;

      begin
         w:=strlen(p);
         for i:=w-1 downto 0 do
           p[i+1]:=p[i];
         p[0]:=chr(w);
         pchar2pstring:=pstring(p);
      end;

    function pstring2pchar(p : pstring) : pchar;

      var
         w : word;
         i : longint;

      begin
         w:=ord(p^[0]);
         for i:=1 to w do
           p^[i-1]:=p^[i];
         p^[w]:=#0;
         pstring2pchar:=pchar(p);
      end;

    function lowercase(c : char) : char;

       begin
          case c of
             #65..#90 : c := chr(ord (c) + 32);
             #154 : c:=#129;  { german }
             #142 : c:=#132;  { german }
             #153 : c:=#148;  { german }
             #144 : c:=#130;  { french }
             #128 : c:=#135;  { french }
             #143 : c:=#134;  { swedish/norge (?) }
             #165 : c:=#164;  { spanish }
             #228 : c:=#229;  { greek }
             #226 : c:=#231;  { greek }
             #232 : c:=#227;  { greek }
          end;
          lowercase := c;
       end;

    function strpnew(const s : string) : pchar;

      var
         p : pchar;

      begin
{ $ifdef CleanUp
         Shouldn't it be nil if s is empty (Pierre M.)?
          I think, that's not a good idea (Florian K.)
        if s = '' then
          begin
          strpnew := Nil;
          end else
          begin
$endif * CleanUp *}
         getmem(p,length(s)+1);
         strpcopy(p,s);
         strpnew:=p;
{ $ifdef CleanUp
          end;
$endif * CleanUp }
      end;

    procedure stringdispose(var p : pstring);

      begin
         if assigned(p) then
           freemem(p,length(p^)+1);
         p:=nil;
      end;

    function stringdup(const s : string) : pstring;

      var
         p : pstring;

      begin
         getmem(p,length(s)+1);
         p^:=s;
         stringdup:=p;
      end;

{****************************************************************************
                           TSTRINGCONTAINER
 ****************************************************************************}

    constructor tstringcontainer.init;

      begin
         root:=nil;
         last:=nil;
         doubles:=true;
      end;

    destructor tstringcontainer.done;

      begin
         clear;
      end;

    procedure tstringcontainer.insert(const s : string);

      var
         hp : pstringitem;

      begin
         if not(doubles) then
           begin
              hp:=root;
              while assigned(hp) do
                begin
                   if hp^.data^=s then exit;
                   hp:=hp^.next;
                end;
           end;
         new(hp);
         hp^.next:=nil;
         hp^.data:=stringdup(s);
         if root=nil then root:=hp
           else last^.next:=hp;
         last:=hp;
      end;

    procedure tstringcontainer.clear;

      var
         hp : pstringitem;

      begin
         hp:=root;
         while assigned(hp) do
           begin
              stringdispose(hp^.data);
              root:=hp^.next;
              dispose(hp);
              hp:=root;
           end;
         last:=nil;
         root:=nil;
      end;

    function tstringcontainer.get : string;

      var
         hs : string;
         hp : pstringitem;

      begin
         if root=nil then hs:=''
           else
           begin
              hs:=root^.data^;
              hp:=root;
              root:=root^.next;
              stringdispose(hp^.data);
              dispose(hp);
           end;
         get:=hs;
      end;

{****************************************************************************
                            TLINKEDLIST_ITEM
 ****************************************************************************}

    destructor tlinkedlist_item.done;

      begin
      end;

{****************************************************************************
                               TLINKEDLIST
 ****************************************************************************}

    constructor tlinkedlist.init;

      begin
         first:=nil;
         last:=nil;
      end;

    destructor tlinkedlist.done;

      begin
         clear;
      end;

    procedure tlinkedlist.clear;

      var
         hp : plinkedlist_item;

      begin
         hp:=first;
         while assigned(hp) do
           begin
              first:=hp^.next;
              dispose(hp,done);
              hp:=first;
           end;
      end;

    procedure tlinkedlist.insertlist(p : plinkedlist);

      begin
         { empty list ? }
         if not(assigned(p^.first)) then
           exit;

         p^.last^.next:=first;

         { we have a double linked list }
         if assigned(first) then
           first^.last:=p^.last;

         first:=p^.first;

         if not(assigned(last)) then
           last:=p^.last;

         { p becomes empty }
         p^.first:=nil;
         p^.last:=nil;
      end;

    procedure tlinkedlist.concat(p : plinkedlist_item);

      begin
         p^.last:=nil;
         p^.next:=nil;
         if not(assigned(first)) then
           first:=p
           else
             begin
                last^.next:=p;
                p^.last:=last;
             end;
         last:=p;
      end;

    procedure tlinkedlist.insert(p : plinkedlist_item);

      begin
         p^.last:=nil;
         p^.next:=nil;
         if not(assigned(first)) then
           last:=p
         else
           begin
              first^.last:=p;
              p^.next:=first;
              first:=p;
           end;
         first:=p;
      end;

    procedure tlinkedlist.concatlist(p : plinkedlist);

     begin
         if not(assigned(p^.first)) then
           exit;

         if not(assigned(first)) then
           first:=p^.first
           else
             begin
                last^.next:=p^.first;
                p^.first^.last:=last;
             end;

         last:=p^.last;

         { make p empty }
         p^.last:=nil;
         p^.first:=nil;
      end;

{****************************************************************************
                               TBUFFEREDFILE
 ****************************************************************************}

    constructor tbufferedfile.init(const filename : string;_bufsize : longint);

      begin
         assign(f,filename);
         bufsize:=_bufsize;
         bufpos:=0;
         getmem(buf,bufsize);
      end;

    destructor tbufferedfile.done;

      begin
         flush;
         system.close(f);
         freemem(buf,bufsize);
      end;

    procedure tbufferedfile.reset;

      begin
         iomode:=1;
         system.reset(f,1);
      end;

    procedure tbufferedfile.rewrite;

      begin
         iomode:=0;
         system.rewrite(f,1);
      end;

    procedure tbufferedfile.flush;

      begin
         if iomode=0 then
           blockwrite(f,buf^,bufpos)
         else
           {!!!!!}
           ;
         bufpos:=0;
      end;

    procedure tbufferedfile.write_string(const s : string);

      begin
        if bufpos+length(s)>bufsize then
          flush;
        move(s[1],(buf+bufpos)^,length(s));
        inc(bufpos,length(s));
      end;

    procedure tbufferedfile.write_pchar(p : pchar);

      var
         l : longint;

      begin
        l:=strlen(p);
        if bufpos+l>bufsize then
          flush;
        move(p^,(buf+bufpos)^,l);
        inc(bufpos,l);
      end;

end.
